/*
 * Object for border traces using PamImage
 * Traces inside components, includes holes.
 * Input image must be strictly black/white.
 * Uses an intermediate 'side_table' containing between-pixel boundaries.
 * (C) Axel Brink, dept. of Artificial Intelligence, University of Groningen
 * October 2007.
 */
 
#ifndef BORDERTRACINGLIB_H
#define BORDERTRACINGLIB_H

#include "pamImage.h"
#include <vector>

// struct for (x, y) pixel positions
struct Pos {
	int x;
	int y;
};

// struct for BorderTrace which is a list of positions inside a big array
struct BorderTrace {
	Pos* list;
	int len;
};

// indicates whether a pixel has a border on its top or left side 
struct PixelSide {
	bool left;
	bool top;
};

class BorderTracing {
	private:
		PamImage* image;            // the image
		Pos* trace_data;            // big array for all border trace data
		Pos* free_data_ptr;         // points to next free position in trace_data
		std::vector<BorderTrace> traces; // vector of all traces
		PixelSide** side_table;     // white/black and black/white transition info for every pixel
		int width;                  // image width
		int height;                 // image height
		void init();                // initialize
		void trace_borders();       // do the thing
		void compute_side_table();  // compute where white/black and black/white transitions occur
		void trace_here(int x, int y); // add trace starting here (a side is above)
		
	public:
		BorderTracing(PamImage* im); // use with .pgm or .pbm that only has black and white
		~BorderTracing();
		std::vector<BorderTrace> getTraces();
		PamImage* draw();
		PamImage* draw_sides();
};

#endif // BORDERTRACINGLIB_H
