/* 
 * This code is the C++ version of the method published on:
 * 
 * Dilip K. Prasad, Chai Quek, Maylor K.H. Leung, and Siu-Yeung Cho, “A parameter independent line fitting method,” 1st IAPR Asian Conference on Pattern Recognition 
 * 
 * Please cite this paper if you use this program
 * 
 ***  This source file is a part of *********
 * 
 * Document feature library (DFLib)
 * 
 * @copyright: Sheng He, University of Groningen (RUG)
 * 
 * Email: heshengxgd@gmail.com
 * 
 *  27, Nov., 2015
 * */

#ifndef _DFLIB_CONTOURLINEFIT_H_
#define _DFLIB_CONTOURLINEFIT_H_


#include <vector>
#include "dflUtils.h"

class dflContourlinefit{
	
	public:
		dflContourlinefit(int verbose=0, bool removeLinear = true,float collinearThresh = 3.0);
		~dflContourlinefit();

		/** input: pixels of contour
		 *  output: the dominant points */
		std::vector<dflPixel> linefit(std::vector<dflPixel>& contour);

	private:		
		void maxlinedev_opt(std::vector<dflPixel>& subcontour,double *max_dev,int *index,
											   double *D_temp,double *precision,double *reliability,double *del_tol_max,double *sum_dev);
		double function_digital_case_with_max_error(double S_max);	
		
		void removeCollinearPoints(std::vector<dflPixel>& points);
		
		bool romoveCollinear;
		float collinearThresh;
		int verbose;
		bool silence;
};	

#endif
